###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

import base64
import enum
import json
import hashlib
import logging

import click

from pywis_pubsub import cli_options
from pywis_pubsub.message import get_link, get_data

LOGGER = logging.getLogger(__name__)


class VerificationMethods(enum.Enum):
    sha256 = 'sha256'
    sha384 = 'sha_384'
    sha512 = 'sha512'
    sha3_256 = 'sha3_256'
    sha3_384 = 'sha3_384'
    sha3_512 = 'sha3_512'


def data_verified(data: bytes, size: int, method: VerificationMethods,
                  value: str):
    """
    Verify integrity of data given a hashing method

    :param data: data to verify
    :param size: `int` size of data
    :param method: method of verification (`VerificationMethods`)
                   default is sha512
    :param value: value of hashing result

    :returns: `bool` of whether data is verified
    """

    LOGGER.debug(f'Verification method is {method}')
    data_value = hashlib.new(method)
    data_value.update(data)
    data_value_base64 = base64.b64encode(data_value.digest()).decode()

    LOGGER.debug('Comparing checksum and data size')

    return data_value_base64 == value and len(data) == size


def verify_data(instance: dict, verify_certs: bool = True) -> bool:
    """
    Validate a JSON instance document against an JSON schema

    :param instance: `dict` of JSON
    :param verify_certs: `bool` whether to verify server
                         certificates (default is `True`)

    :returns: `bool` of verification result
    """

    if 'integrity' not in instance['properties']:
        msg = 'No properties.integrity set. Exiting'
        LOGGER.error(msg)
        raise ValueError(msg)

    try:
        LOGGER.debug('Downloading data')
        data = get_data(instance, verify_certs)
    except Exception as err:
        LOGGER.error(err)
        raise ValueError(err)

    method = instance['properties']['integrity']['method']
    LOGGER.debug(f'method: {method}')
    value = instance['properties']['integrity']['value']
    LOGGER.debug(f'value: {value}')

    if 'content' in instance['properties']:
        size = instance['properties']['content']['size']
    else:
        size = get_link(instance['links'])['length']

    LOGGER.debug(f'size: {size}')
    return data_verified(data, size, method, value)


@click.command()
@click.pass_context
@click.argument('message', type=click.File())
@click.option('--verify-certs', '-vc', is_flag=True, default=True,
              help='Whether to verify server certificates')
@cli_options.OPTION_VERBOSITY
def verify(ctx, message, verify_certs, verbosity):
    """Verify a message"""

    try:
        message_dict = json.load(message)
    except json.decoder.JSONDecodeError as err:
        raise click.ClickException(f'Malformed message: {err}')

    verifies = verify_data(message_dict)

    if not verifies:
        raise click.ClickException('Verification failed')
    else:
        click.echo('Valid message')



###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

import base64
from datetime import datetime
from enum import Enum
import hashlib
import json
import logging
import mimetypes
from typing import Union
import uuid

import click
import requests

from pywis_pubsub import cli_options
from pywis_pubsub import util
from pywis_pubsub.message import LINK_TYPES
from pywis_pubsub.mqtt import MQTTPubSubClient
from pywis_pubsub.ets import validate


LOGGER = logging.getLogger(__name__)


class SecureHashAlgorithms(Enum):
    sha256 = 'sha256'
    sha384 = 'sha_384'
    sha512 = 'sha512'
    sha3_256 = 'sha3_256'
    sha3_384 = 'sha3_384'
    sha3_512 = 'sha3_512'


def generate_checksum(data: bytes, algorithm: SecureHashAlgorithms) -> str:
    """
    Generate a checksum of message file

    :param data: bytes of data
    :param algorithm: secure hash algorithm (SecureHashAlgorithm)

    :returns: hexdigest
    """

    sh = getattr(hashlib, algorithm)()
    sh.update(data)

    b64_digest = base64.b64encode(sh.digest()).decode()
    LOGGER.debug(f'Base 64 encoded digest: {b64_digest}')
    return b64_digest


def get_file_info(public_data_url: str) -> dict:
    """
    get filename, length and calculate checksum from public URL

    :param public_data_url: `str` defining publicly accessible URL

    :returns: `dict` of file information
    """

    res = requests.get(public_data_url)
    # raise HTTPError, if on occurred:
    res.raise_for_status()

    filebytes = res.content
    checksum_type = SecureHashAlgorithms.sha512.value
    file_info = {
        'filename': public_data_url.split('/')[-1],
        'checksum_value': generate_checksum(filebytes, checksum_type),
        'checksum_type': checksum_type,
        'size': len(filebytes)
    }

    if len(filebytes) < 4096:
        file_info['data'] = base64.b64encode(filebytes)

    return file_info


def create_message(topic: str, content_type: str, url: str, identifier: str,
                   inline: bool = False, geometry: list = [],
                   datetime_: datetime = None,
                   start_datetime: datetime = None,
                   end_datetime: datetime = None,
                   metadata_id: str = None,
                   wigos_station_identifier: str = None,
                   operation: Union[str] = 'create') -> dict:
    """
    Create WIS2 compliant message

    :param topic: `str` of topic
    :param url: `str` of url pointing to data
    :param identifier: `str` of unique-id to help global broker deduplicate data  # noqa
    :param inline: `bool` of whether to publish the data inline as base64
             (default False)
    :param datetime_: `datetime` object of data (time instant)
    :param start_datetime: `datetime` object of start date
    :param end_datetime: `datetime` object of end date
    :param geometry: point array defining longitude,latitude,elevation
               (elevation is optional)
    :param metadata_id: `str` of WCMP2 metadata record identifier
    :param wigos_station_identifier: `str` of WSI for station as used in OSCAR
    :param operation: `str` of message operation

    :returns: `dict` of message
    """

    publish_datetime = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')

    # get filename, length and calculate checksum
    # raises HTTPError if file can not be accessed
    file_info = get_file_info(url)

    if geometry:
        geometry2 = {
            'type': 'Point',
            'coordinates': geometry
        }
    else:
        geometry2 = None

    if content_type is None:
        content_type2 = mimetypes.guess_type(url)[0]
        if content_type2 is None:
            LOGGER.warning('Unknown content type')
            content_type2 = 'application/octet-stream'
    else:
        content_type2 = content_type

    message = {
            'id': identifier,
            'type': 'Feature',
            'conformsTo': ['http://wis.wmo.int/spec/wnm/1/conf/core'],
            'geometry': geometry2,
            'properties': {
                'data_id': f"{topic}/{file_info['filename']}",
                'pubtime': publish_datetime,
                'integrity': {
                    'method': file_info['checksum_type'],
                    'value': file_info['checksum_value']
                },
            },
            'links': [{
                'rel': LINK_TYPES[operation],
                'type': content_type2,
                'href': url,
                'length': file_info['size']
            }]
    }

    if None not in [start_datetime, end_datetime]:
        LOGGER.debug('Setting time extent')
        message['properties']['start_datetime'] = start_datetime
        message['properties']['endstart_datetime'] = end_datetime
    else:
        LOGGER.debug('Setting time instant')
        message['properties']['datetime'] = datetime_

    if file_info.get('data') is not None and inline:
        LOGGER.debug('Including data inline via properties.content')
        message['properties']['content'] = {
            'encoding': 'base64',
            'value': file_info['data'],
            'size': file_info['size']
        }

    if metadata_id is not None:
        message['properties']['metadata_id'] = metadata_id

    if wigos_station_identifier is not None:
        message['properties']['wigos_station_identifier'] = wigos_station_identifier  # noqa

    return message


@click.command()
@click.pass_context
@cli_options.OPTION_CONFIG
@cli_options.OPTION_VERBOSITY
@click.option('--wnm', '-wnm', type=click.File(), help='url of data')
@click.option('--url', '-u', help='url of data')
@click.option('--inline', '-in', default=False,
              help='whether to publish the data inline as base64 (default=False)')  # noqa
@click.option('--datetime', '-d', 'datetime_',
              help='Datetime instant or extent')
@click.option('--topic', '-t', help='topic to publish to')
@click.option('--geometry', '-g',
              help='point geometry as longitude,latitude,elevation (elevation is optional)')  # noqa
@click.option('--metadata-id', '-m', help='WCMP2 metadata record identifier')
@click.option('--wigos_station_identifier', '-w',
              help='WIGOS station identifier')
@click.option('--operation', '-op', type=click.Choice(LINK_TYPES.keys()),
              default='create', help='message operation')
def publish(ctx, wnm, config, url, topic, datetime_,
            inline=False, geometry=[], metadata_id=None,
            wigos_station_identifier=None, operation='create',
            verbosity='NOTSET'):
    """Publish a WIS2 Notification Message"""

    if config is None:
        raise click.ClickException('missing -c/--config')

    if wnm is None and url is None:
        raise click.ClickException('missing required arguments')

    config = util.yaml_load(config)

    datetime_2 = None
    start_datetime = None
    end_datetime = None

    broker = config.get('broker')
    qos = int(config.get('qos', 1))

    if topic is None:
        topic2 = config.get('publish_topic')
    else:
        topic2 = topic

    if wnm is not None:
        if config.get('validate_message', False):
            ctx.invoke(validate, message=wnm)
            wnm.seek(0)
        message = json.load(wnm)
    else:
        if datetime_ is not None:
            if '/' in datetime_:
                start, end = datetime_.split('/')
                if start:
                    start_datetime = datetime.strptime(
                        start, '%Y-%m-%dT%H:%M:%SZ')
                if end:
                    end_datetime = datetime.strptime(
                        end, '%Y-%m-%dT%H:%M:%SZ')
            else:
                datetime_2 = datetime.strptime(
                    datetime_, '%Y-%m-%dT%H:%M:%SZ')

        message = create_message(
            topic=topic2,
            content_type=config.get('content_type'),
            url=url,
            identifier=str(uuid.uuid4()),
            inline=inline,
            datetime_=datetime_2,
            start_datetime=start_datetime,
            end_datetime=end_datetime,
            geometry=geometry,
            metadata_id=metadata_id,
            wigos_station_identifier=wigos_station_identifier,
            operation=operation
        )

    client = MQTTPubSubClient(broker)
    click.echo(f'Connected to broker {client.broker_safe_url}')
    click.echo(f'Publishing message to topic={topic2}')
    client.pub(topic2, json.dumps(message, default=util.json_serial), qos)



###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

# executable test suite as per WNM, Annex A

import json
import logging
from uuid import UUID

import click
from jsonschema.validators import Draft202012Validator


import pywis_pubsub
from pywis_pubsub.errors import TestSuiteError
from pywis_pubsub.schema import MESSAGE_SCHEMA
from pywis_pubsub.message import get_link
from pywis_pubsub.util import (get_cli_common_options,
                               get_current_datetime_rfc3339, urlopen_)

LOGGER = logging.getLogger(__name__)


def gen_test_id(test_id: str) -> str:
    """
    Convenience function to print test identifier as URI

    :param test_id: test suite identifier

    :returns: test identifier as URI
    """

    return f'http://wis.wmo.int/spec/wnm/1/conf/core/{test_id}'


class WNMTestSuite:
    """Test suite for WIS Notification Message"""

    def __init__(self, data: dict):
        """
        initializer

        :param data: dict of WNM JSON

        :returns: `pywis_pubsub.ets.WNMTestSuite`
        """

        self.test_id = None
        self.message = data
        self.errors = []
        self.report = []

    def run_tests(self, fail_on_schema_validation=False):
        """Convenience function to run all tests"""

        results = []
        tests = []

        ets_report = {
            'summary': {},
            'generated_by': f'pywis-pubsub {pywis_pubsub.__version__} (https://github.com/World-Meteorological-Organization/pywis-pubsub)'  # noqa
        }

        for f in dir(WNMTestSuite):
            if all([
                    callable(getattr(WNMTestSuite, f)),
                    f.startswith('test_requirement'),
                    not f.endswith('validation')]):

                tests.append(f)

        validation_result = self.test_requirement_validation()
        if validation_result['code'] == 'FAILED':
            if fail_on_schema_validation:
                msg = ('Record fails WNM validation. Stopping ETS ',
                       f"errors: {validation_result['errors']}")
                LOGGER.error(msg)
                raise ValueError(msg)

        for t in tests:
            result = getattr(self, t)()
            results.append(result)
            if result['code'] == 'FAILED':
                self.errors.append(result)

        for code in ['PASSED', 'FAILED', 'SKIPPED']:
            r = len([t for t in results if t['code'] == code])
            ets_report['summary'][code] = r

        ets_report['tests'] = results
        ets_report['datetime'] = get_current_datetime_rfc3339()

        return {
            'ets-report': ets_report
        }

    def raise_for_status(self):
        """
        Raise error if one or more failures were found during validation.

        :returns: `pywcmp.errors.TestSuiteError` or `None`
        """

        if len(self.errors) > 0:
            raise TestSuiteError('Invalid WNM', self.errors)

    def test_requirement_validation(self):
        """
        Validate that a WNM is valid to the authoritative WNM schema.
        """

        validation_errors = []

        status = {
            'id': gen_test_id('validation'),
            'code': 'PASSED'
        }

        if not MESSAGE_SCHEMA.exists():
            msg = "WNM schema missing. Run 'pywis-pubsub schema sync' to cache"
            LOGGER.error(msg)
            raise RuntimeError(msg)

        with MESSAGE_SCHEMA.open() as fh:
            LOGGER.debug(f'Validating {self.message} against {MESSAGE_SCHEMA}')
            validator = Draft202012Validator(json.load(fh))

            for error in validator.iter_errors(self.message):
                LOGGER.debug(f'{error.json_path}: {error.message}')
                validation_errors.append(f'{error.json_path}: {error.message}')

            if validation_errors:
                status['code'] = 'FAILED'
                status['message'] = f'{len(validation_errors)} error(s)'
                status['errors'] = validation_errors

        return status

    def test_requirement_id(self):
        """
        Check for the existence of a valid id property.
        """

        status = {
            'id': gen_test_id('id'),
            'code': 'PASSED',
        }

        try:
            UUID(self.message['id'])
        except ValueError as err:
            status['code'] = 'FAILED'
            status['message'] = f'Invalid UUID: {err}'

        return status

    def test_requirement_geometry(self):
        """
        Check for the existence of a valid geometry property.
        """

        status = {
            'id': gen_test_id('geometry'),
            'code': 'PASSED',
            'message': 'Passes given document is compliant/valid'
        }

        return status

    def test_requirement_temporal(self):
        """
        Check for the existence of a valid datetime property/properties.
        """

        status = {
            'id': gen_test_id('temporal'),
            'code': 'PASSED',
        }

        for dtp in ['datetime', 'start_datetime', 'end_datetime']:
            value = self.message['properties'].get(dtp)
            if value is not None:
                if not value.endswith(('Z', '+00:00')):
                    status['code'] = 'FAILED'
                    status['message'] = 'date-time not in UTC'

        return status

    def test_requirement_data_id(self):
        """
        Check for the existence of a valid properties.data_id.
        """

        status = {
            'id': gen_test_id('data_id'),
            'code': 'PASSED',
            'message': 'Passes given document is compliant/valid'
        }

        return status

    def test_requirement_version(self):
        """
        Validate that a WNM provides a valid properties.version.
        """

        status = {
            'id': gen_test_id('version'),
            'code': 'PASSED',
            'message': 'Passes given document is compliant/valid'
        }

        return status

    def test_requirement_conformance(self):
        """
        Validate that a WCMP record provides valid conformance information.
        """

        status = {
            'id': gen_test_id('conformance'),
            'code': 'PASSED',
            'message': 'Passes given document is compliant/valid'
        }

        return status

    def test_requirement_links(self):
        """
        Validate that a WNM provides valid link property/properties.
        """

        status = {
            'id': gen_test_id('links'),
            'code': 'PASSED'
        }

        links = self.message['links']

        LOGGER.debug('Checking for at least one link')
        if len(links) < 1:
            status['code'] = 'FAILED'
            status['message'] = 'missing at least one link'
            return status

        for link in links:
            LOGGER.debug('Checking for required protocols')
            if not link['href'].startswith(('http', 'https', 'ftp', 'sftp')):
                status['code'] = 'FAILED'
                status['message'] = 'Invalid link protocol'
                return status

        if not get_link(links):
            status['code'] = 'FAILED'
            status['message'] = 'Missing required link relations'
            return status

        return status


@click.group()
def ets():
    """executable test suite"""
    pass


@click.command()
@click.pass_context
@get_cli_common_options
@click.argument('file_or_url')
@click.option('--fail-on-schema-validation/--no-fail-on-schema-validation',
              '-f', default=True,
              help='Stop the ETS on failing schema validation')
def validate(ctx, file_or_url, logfile, verbosity,
             fail_on_schema_validation=True):
    """validate against the abstract test suite"""

    click.echo(f'Opening {file_or_url}')

    if file_or_url.startswith('http'):
        content = urlopen_(file_or_url).read()
    else:
        with open(file_or_url) as fh:
            content = fh.read()

    content = json.loads(content)
    click.echo(f'Validating {file_or_url}')

    ts = WNMTestSuite(content)
    try:
        results = ts.run_tests(fail_on_schema_validation)
    except Exception as err:
        raise click.ClickException(err)

    click.echo(json.dumps(results, indent=4))
    ctx.exit(results['ets-report']['summary']['FAILED'])


ets.add_command(validate)



###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

import json
import logging
from pathlib import Path
import random

import click

from pywis_pubsub import cli_options
from pywis_pubsub import util
from pywis_pubsub.ets import WNMTestSuite
from pywis_pubsub.geometry import is_message_within_bbox
from pywis_pubsub.hook import load_hook
from pywis_pubsub.message import get_link, get_data
from pywis_pubsub.mqtt import MQTTPubSubClient
from pywis_pubsub.storage import STORAGES
from pywis_pubsub.verification import data_verified


LOGGER = logging.getLogger(__name__)


def on_message_handler(client, userdata, msg):
    """message handler"""

    LOGGER.info(f'Topic: {msg.topic}')
    LOGGER.info(f'Message:\n{msg.payload}')

    msg_dict = json.loads(msg.payload)

    try:
        if userdata.get('validate_message', False):
            LOGGER.debug('Validating message')

            ts = WNMTestSuite(msg_dict)
            _ = ts.run_tests(fail_on_schema_validation=True)

    except Exception as err:
        LOGGER.error(f'Cannot validate message: {err}')
        return

    if userdata.get('bbox') and msg_dict.get('geometry') is not None:
        LOGGER.debug('Performing spatial filtering')
        if not bool(msg_dict['geometry']):
            LOGGER.error(f"Invalid geometry: {msg_dict['geometry']}")
            return
        if is_message_within_bbox(msg_dict['geometry'], userdata['bbox']):
            LOGGER.debug('Message geometry is within bbox')
        else:
            LOGGER.debug('Message geometry not within bbox; skipping')
            return

    clink = get_link(msg_dict['links'])
    if not clink:
        LOGGER.warning('No valid data link found')
        return

    LOGGER.info(f"Received message with data URL: {clink.get('href')}")

    if userdata.get('storage') is not None:
        LOGGER.debug('Saving data')
        try:
            LOGGER.debug('Downloading data')
            data = get_data(msg_dict, userdata.get('verify_certs'))
        except Exception as err:
            LOGGER.error(err)
            return
        if ('integrity' in msg_dict['properties'] and
                userdata.get('verify_data', True)):
            LOGGER.debug('Verifying data')

            method = msg_dict['properties']['integrity']['method']
            value = msg_dict['properties']['integrity']['value']
            if 'content' in msg_dict['properties']:
                size = msg_dict['properties']['content']['size']
            else:
                size = clink['length']

            LOGGER.debug(method)
            if not data_verified(data, size, method, value):
                LOGGER.error('Data verification failed; not saving')
                return
            else:
                LOGGER.debug('Data verification passed')

        filepath = userdata['storage']['options'].get('filepath', 'data_id')
        LOGGER.debug(f'Using {filepath} for naming filepath')

        link = get_link(msg_dict['links'])

        if filepath == 'link':
            LOGGER.debug('Using link as filepath')
            # fetch link and use local path, stripping slashes
            filename = link['href'].split('/', 3)[-1].strip('/')
        elif filepath == 'combined':
            LOGGER.debug('Using combined data_id+link extension as filepath')
            filename = msg_dict['properties']['data_id']
            suffix = Path(link).suffix
            if suffix != '':
                LOGGER.debug(f'File extension found: {suffix}')
                filename = f'{filename}{suffix}'
            else:
                LOGGER.debug('File extension not found. Trying media type')
                media_type = link.get('type')
                if media_type is not None:
                    suffix = util.guess_extension(media_type)
                    if suffix is not None:
                        filename = f'{filename}{suffix}'
                    else:
                        LOGGER.debug('No extension found. Giving up / using data_id')  # noqa
                else:
                    LOGGER.debug('No media type found. Giving up / using data_id')  # noqa
        else:
            LOGGER.debug('Using data_id as filepath')
            filename = msg_dict['properties'].get('data_id')
            if filename is None:
                LOGGER.error('no data_id found')
                return

        LOGGER.debug(f'filename: {filename}')

        content_type = link.get('type', 'application/octet-stream')

        storage_class = STORAGES[userdata.get('storage').get('type')]
        storage_object = storage_class(userdata['storage'])

        if not msg_dict['properties'].get('cache', True):
            LOGGER.debug(f'No caching requested; not saving {filename}')
            return

        if link.get('rel') == 'deletion':
            LOGGER.debug('Delete specified')
            storage_object.delete(filename)
        elif link.get('rel') == 'update':
            LOGGER.debug('Update specified')
            storage_object.save(data, filename, content_type)
        else:
            if storage_object.exists(filename):
                LOGGER.debug('Duplicate detected; not saving')
                return

            LOGGER.debug('Saving')
            storage_object.save(data, filename, content_type)

    if userdata.get('hook') is not None:
        LOGGER.debug(f"Hook detected: {userdata['hook']}")
        try:
            hook = load_hook(userdata['hook'])
            LOGGER.debug('Executing hook')
            hook.execute(msg.topic, msg_dict)
        except Exception as err:
            msg = f'Hook failed: {err}'
            LOGGER.error(msg, exc_info=True)


@click.command()
@click.pass_context
@cli_options.OPTION_CONFIG
@cli_options.OPTION_VERBOSITY
@click.option('--bbox', '-b', help='Bounding box filter')
@click.option('--download', '-d', is_flag=True, help='Download data')
def subscribe(ctx, config, download, bbox=[], verbosity='NOTSET'):
    """Subscribe to a broker/topic and optionally download data"""

    if config is None:
        raise click.ClickException('missing --config')
    config = util.yaml_load(config)

    broker = config.get('broker')
    client_id_prefix = config.get('client_id', 'pywis-pubsub')
    client_id = f'{client_id_prefix}-{random.randint(0, 1000)}'
    qos = int(config.get('qos', 1))
    subscribe_topics = config.get('subscribe_topics', [])
    verify_certs = config.get('verify_certs', True)
    certfile = config.get('certfile')
    keyfile = config.get('keyfile')

    options = {
        'verify_certs': verify_certs,
        'certfile': certfile,
        'keyfile': keyfile
    }
    options['client_id'] = client_id
    options['clean_session'] = config.get('clean_session', True)

    if bbox:
        options['bbox'] = [float(i) for i in bbox.split(',')]

    if download:
        options['storage'] = config['storage']

    options['verify_data'] = config.get('verify_data', True)
    options['validate_message'] = config.get('validate_message', False)
    options['hook'] = config.get('hook')

    client = MQTTPubSubClient(broker, options)
    client.bind('on_message', on_message_handler)
    click.echo(f'Connected to broker {client.broker_safe_url}')
    click.echo(f'Subscribing to subscribe_topics {subscribe_topics}')
    client.sub(subscribe_topics, qos)



###############################################################################
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
###############################################################################

from abc import ABC, abstractmethod
from datetime import datetime
import logging
from pathlib import Path
import shutil
from typing import Iterator, Tuple

LOGGER = logging.getLogger(__name__)


class Storage(ABC):
    # @abstractmethod
    def __init__(self, defs):
        self.type = defs.get('type')
        self.options = defs.get('options')

    @abstractmethod
    def setup(self) -> bool:
        """
        Setup harness

        :returns: `bool` of setup result
        """

        raise NotImplementedError()

    @abstractmethod
    def teardown(self) -> bool:
        """
        Teardown setup

        :returns: `bool` of setup result
        """

        raise NotImplementedError()

    @abstractmethod
    def exists(self, filename: Path) -> bool:
        """
        Verify whether data already exists

        :param filename: `Path` of storage object/file

        :returns: `bool` of whether the filepath exists in storage
        """

        raise NotImplementedError()

    @abstractmethod
    def list_contents_by_date(self, basepath: str = None,
                             recursive: bool = False) -> Iterator[Tuple[str, datetime]]:  # noqa
        """
        List storage paths starting

        :param basepath: basepath
        :param recursive: whether to list recursively (default=`False`)

        :returns: `generator` of contents
        """

        raise NotImplementedError()

    @abstractmethod
    def save(self, data: bytes, filename: Path,
             content_type: str = 'application/octet-stream') -> bool:
        """
        Save data to storage

        :param data: `bytes` of data
        :param filename: `str` of filename
        :param content_type: media type (default is `application/octet-stream`)

        :returns: `bool` of save result
        """

        raise NotImplementedError()

    @abstractmethod
    def delete(self, filename: Path) -> bool:
        """
        Deletes data from storage

        :param filename: `str` of filename

        :returns: `bool` of delete result
        """

        raise NotImplementedError()


class FileSystem(Storage):
    def setup(self) -> bool:

        basedir = Path(self.options['basedir'])
        LOGGER.debug(f'Creating directory {basedir}')
        basedir.mkdir(parents=True, exist_ok=True)

        return True

    def teardown(self) -> bool:
        LOGGER.debug(f"Deleting directory {self.options['basedir']}")
        shutil.rmtree(self.options['basedir'])

        return True

    def exists(self, filename: Path) -> bool:

        filepath = Path(self.options['basedir']) / filename

        return filepath.exists()

    def list_contents_by_date(self, basepath: str = None,
                              recursive: bool = False) -> Iterator[Tuple[Path, datetime]]:  # noqa

        if recursive:
            func = 'rglob'
        if recursive:
            func = 'glob'

        if basepath is None:
            basepath = '/'

        for p in getattr(Path(basepath), func)('*'):
            if p.is_file():
                yield p, datetime.fromtimestamp(p.stat().st_mtime)

    def save(self, data: bytes, filename: Path,
             content_type: str = 'application/octet-stream') -> bool:

        filepath = Path(self.options['basedir']) / filename

        LOGGER.debug(f'Creating directory {filepath.parent}')
        filepath.parent.mkdir(parents=True, exist_ok=True)

        LOGGER.debug(f'Saving data to {filepath}')
        with filepath.open('wb') as fh:
            fh.write(data)

        LOGGER.info(f'Data saved to {filepath}')

        return True

    def delete(self, filename: Path) -> bool:

        filepath = Path(self.options['basedir']) / filename

        LOGGER.debug(f'Deleting file {filepath}')
        filepath.unlink()

        LOGGER.info(f'Deleted file {filepath}')

        return True


class S3(Storage):

    @staticmethod
    def _get_client(self):

        import boto3

        s3_url = self.options['url']
        self.s3_bucket = self.options['bucket']

        s3_client = boto3.client('s3', endpoint_url=s3_url)

        return s3_client

    def teardown(self) -> bool:

        s3_client = self._get_client(self)

        try:
            objects_to_delete = s3_client.list_objects(
                Bucket=self.s3_bucket, Prefix='')
            objs = [{'Key' : k} for k in [obj['Key'] for obj in objects_to_delete.get('Contents', [])]]  # noqa
            delete_keys = {'Objects': objs}

            LOGGER.debug('Deleting all bucket objects')
            s3_client.delete_objects(Bucket=self.s3_bucket, Delete=delete_keys)

            LOGGER.debug(f'Deleting bucket {self.s3_bucket}')
            s3_client.delete_bucket(Bucket=self.s3_bucket)
        except Exception as err:
            LOGGER.error(err)
            return False

        return True

    def exists(self, filename: Path) -> bool:

        s3_client = self._get_client(self)

        try:
            s3_client.head_object(Bucket=self.s3_bucket, Key=filename)
        except Exception:
            return False

        return True

    def setup(self) -> bool:

        s3_client = self._get_client(self)

        try:
            LOGGER.debug(f'Creating bucket {self.s3_bucket}')
            s3_client.create_bucket(Bucket=self.s3_bucket)
        except Exception as err:
            LOGGER.error(err)
            return False

        return True

    def list_contents_by_date(self, basepath: str = None,
                              recursive: bool = False) -> Iterator[Tuple[Path, datetime]]:  # noqa

        s3_client = self._get_client(self)

        params = {
            'Bucket': self.s3_bucket
        }

        if basepath is not None:
            params['Prefix'] = basepath

        objects = s3_client.list_objects(**params)

        for obj in objects['Contents']:
            yield obj['Key'], obj['LastModified']

    def save(self, data: bytes, filename: Path,
             content_type: str = 'application/octet-stream') -> bool:

        s3_client = self._get_client(self)

        try:
            s3_client.put_object(Body=data, Bucket=self.s3_bucket,
                                 Key=filename, ContentType=content_type)
        except Exception as err:
            LOGGER.error(err)
            return False

        LOGGER.info(f'Data saved to {filename}')

        return True

    def delete(self, filename: Path) -> bool:

        s3_client = self._get_client(self)

        LOGGER.debug(f'Deleting object {filename}')

        try:
            s3_client.delete_object(Bucket=self.s3_bucket, Key=filename)
        except Exception as err:
            LOGGER.error(err)
            return False

        LOGGER.info(f'Deleted object {filename}')

        return True


STORAGES = {
    'fs': FileSystem,
    'S3': S3
}
